!
! stub header for disabled mpi 
!
       INTEGER MPI_COMM_WORLD
       PARAMETER (MPI_COMM_WORLD=1)
       INTEGER MPI_INTEGER
       PARAMETER (MPI_INTEGER=1)
       INTEGER MPI_STATUS_SIZE
       PARAMETER (MPI_STATUS_SIZE=1)
       INTEGER MPI_ANY_SOURCE
       PARAMETER (MPI_ANY_SOURCE=1)
       INTEGER MPI_SOURCE
       PARAMETER (MPI_SOURCE=1)
       INTEGER MPI_DOUBLE_COMPLEX
       PARAMETER (MPI_DOUBLE_COMPLEX=1)
       INTEGER MPI_DOUBLE_PRECISION
       PARAMETER (MPI_DOUBLE_PRECISION=1)
       INTEGER MPI_BSEND_OVERHEAD
       INTEGER MPI_MIN
       PARAMETER (MPI_MIN=1)
       INTEGER MPI_MAX
       PARAMETER (MPI_MAX=1)
       INTEGER MPI_SUM
       PARAMETER (MPI_SUM=1)
       INTEGER MPI_LOGICAL
       PARAMETER (MPI_LOGICAL=1)
       INTEGER MPI_LAND
       PARAMETER (MPI_LAND=1)
       INTEGER MPI_LOR
       PARAMETER (MPI_LOR=1)
